@@ -43,12 +43,42 @@
 #include "tlshd.h"
 #include "netlink.h"
 
+static int tlshd_client_get_truststore(gnutls_certificate_credentials_t cred)
+{
+	char *pathname;
+	int ret;
+
+	if (tlshd_config_get_client_truststore(&pathname)) {
+		ret = gnutls_certificate_set_x509_trust_file(cred, pathname,
+							     GNUTLS_X509_FMT_PEM);
+		free(pathname);
+	} else
+		ret = gnutls_certificate_set_x509_system_trust(cred);
+	if (ret < 0) {
+		tlshd_log_gnutls_error(ret);
+		return ret;
+	}
+	tlshd_log_debug("System trust: Loaded %d certificate(s).", ret);
+
+	if (tlshd_config_get_client_crl(&pathname)) {
+		ret = gnutls_certificate_set_x509_crl_file(cred, pathname,
+							   GNUTLS_X509_FMT_PEM);
+		free(pathname);
+		if (ret < 0 ) {
+			tlshd_log_gnutls_error(ret);
+			return ret;
+		}
+		tlshd_log_debug("System CRL: Loaded %d CRL(s).", ret);
+	}
+
+	return GNUTLS_E_SUCCESS;
+}
+
 static void tlshd_tls13_client_anon_handshake(struct tlshd_handshake_parms *parms)
 {
 	gnutls_certificate_credentials_t xcred;
 	gnutls_session_t session;
 	unsigned int flags;
-	char *cafile;
 	int ret;
 
 	ret = gnutls_certificate_allocate_credentials(&xcred);
@@ -65,17 +95,9 @@ static void tlshd_tls13_client_anon_handshake(struct tlshd_handshake_parms *parm
 	gnutls_certificate_set_flags(xcred,
 			GNUTLS_CERTIFICATE_SKIP_KEY_CERT_MATCH | GNUTLS_CERTIFICATE_SKIP_OCSP_RESPONSE_CHECK);
 
-	if (tlshd_config_get_client_truststore(&cafile)) {
-		ret = gnutls_certificate_set_x509_trust_file(xcred, cafile,
-							     GNUTLS_X509_FMT_PEM);
-		free(cafile);
-	} else
-		ret = gnutls_certificate_set_x509_system_trust(xcred);
-	if (ret < 0) {
-		tlshd_log_gnutls_error(ret);
+	ret = tlshd_client_get_truststore(xcred);
+	if (ret != GNUTLS_E_SUCCESS)
 		goto out_free_creds;
-	}
-	tlshd_log_debug("System trust: Loaded %d certificate(s).", ret);
 
 	flags = GNUTLS_CLIENT;
 	ret = gnutls_init(&session, flags);
@@ -116,9 +138,6 @@ static gnutls_privkey_t tlshd_privkey;
 static unsigned int tlshd_certs_len = TLSHD_MAX_CERTS;
 static gnutls_pcert_st tlshd_certs[TLSHD_MAX_CERTS];
 
-/*
- * XXX: After this point, tlshd_certs should be deinited on error.
- */
 static bool tlshd_x509_client_get_certs(struct tlshd_handshake_parms *parms)
 {
 	if (parms->x509_cert != TLS_NO_CERT)
@@ -127,9 +146,14 @@ static bool tlshd_x509_client_get_certs(struct tlshd_handshake_parms *parms)
 	return tlshd_config_get_client_certs(tlshd_certs, &tlshd_certs_len);
 }
 
-/*
- * XXX: After this point, tlshd_privkey should be deinited on error.
- */
+static void tlshd_x509_client_put_certs(void)
+{
+	unsigned int i;
+
+	for (i = 0; i < tlshd_certs_len; i++)
+		gnutls_pcert_deinit(&tlshd_certs[i]);
+}
+
 static bool tlshd_x509_client_get_privkey(struct tlshd_handshake_parms *parms)
 {
 	if (parms->x509_privkey != TLS_NO_PRIVKEY)
@@ -138,6 +162,11 @@ static bool tlshd_x509_client_get_privkey(struct tlshd_handshake_parms *parms)
 	return tlshd_config_get_client_privkey(&tlshd_privkey);
 }
 
+static void tlshd_x509_client_put_privkey(void)
+{
+	gnutls_privkey_deinit(tlshd_privkey);
+}
+
 static void tlshd_x509_log_issuers(const gnutls_datum_t *req_ca_rdn, int nreqs)
 {
 	char issuer_dn[256];
@@ -207,9 +236,7 @@ static int tlshd_client_x509_verify_function(gnutls_session_t session,
 					     struct tlshd_handshake_parms *parms)
 {
 	const gnutls_datum_t *peercerts;
-	gnutls_certificate_type_t type;
-	unsigned int i, status;
-	gnutls_datum_t out;
+	unsigned int i, status, num_peercerts;
 	int ret;
 
 	ret = gnutls_certificate_verify_peers3(session, parms->peername,
@@ -218,12 +245,6 @@ static int tlshd_client_x509_verify_function(gnutls_session_t session,
 		tlshd_log_gnutls_error(ret);
 		return GNUTLS_E_CERTIFICATE_ERROR;
 	}
-
-        type = gnutls_certificate_type_get(session);
-        gnutls_certificate_verification_status_print(status, type, &out, 0);
-        tlshd_log_debug("%s", out.data);
-        gnutls_free(out.data);
-
         if (status)
                 return GNUTLS_E_CERTIFICATE_ERROR;
 
@@ -231,20 +252,18 @@ static int tlshd_client_x509_verify_function(gnutls_session_t session,
 	 * to get picky. Kernel would have to tell us what to look for
 	 * via a netlink attribute. */
 
-	peercerts = gnutls_certificate_get_peers(session,
-						 &parms->num_remote_peerids);
-	if (!peercerts || parms->num_remote_peerids == 0) {
+	peercerts = gnutls_certificate_get_peers(session, &num_peercerts);
+	if (!peercerts || num_peercerts == 0) {
 		tlshd_log_debug("The peer cert list is empty.\n");
                 return GNUTLS_E_CERTIFICATE_ERROR;
 	}
 
 	tlshd_log_debug("The peer offered %u certificate(s).\n",
-			parms->num_remote_peerids);
+			num_peercerts);
 
-	if (parms->num_remote_peerids > ARRAY_SIZE(parms->remote_peerid))
-		parms->num_remote_peerids = ARRAY_SIZE(parms->remote_peerid);
-	for (i = 0; i < parms->num_remote_peerids; i++) {
+	for (i = 0; i < num_peercerts; i++) {
 		gnutls_x509_crt_t cert;
+		key_serial_t peerid;
 
 		gnutls_x509_crt_init(&cert);
 		ret = gnutls_x509_crt_import(cert, &peercerts[i],
@@ -254,8 +273,8 @@ static int tlshd_client_x509_verify_function(gnutls_session_t session,
 			gnutls_x509_crt_deinit(cert);
 			return GNUTLS_E_CERTIFICATE_ERROR;
 		}
-		parms->remote_peerid[i] =
-			tlshd_keyring_create_cert(cert, parms->peername);
+		peerid = tlshd_keyring_create_cert(cert, parms->peername);
+		g_array_append_val(parms->remote_peerids, peerid);
 		gnutls_x509_crt_deinit(cert);
 	}
 
@@ -274,7 +293,6 @@ static void tlshd_tls13_client_x509_handshake(struct tlshd_handshake_parms *parm
 	gnutls_certificate_credentials_t xcred;
 	gnutls_session_t session;
 	unsigned int flags;
-	char *cafile;
 	int ret;
 
 	ret = gnutls_certificate_allocate_credentials(&xcred);
@@ -283,22 +301,14 @@ static void tlshd_tls13_client_x509_handshake(struct tlshd_handshake_parms *parm
 		return;
 	}
 
-	if (tlshd_config_get_client_truststore(&cafile)) {
-		ret = gnutls_certificate_set_x509_trust_file(xcred, cafile,
-							     GNUTLS_X509_FMT_PEM);
-		free(cafile);
-	} else
-		ret = gnutls_certificate_set_x509_system_trust(xcred);
-	if (ret < 0) {
-		tlshd_log_gnutls_error(ret);
+	ret = tlshd_client_get_truststore(xcred);
+	if (ret != GNUTLS_E_SUCCESS)
 		goto out_free_creds;
-	}
-	tlshd_log_debug("System trust: Loaded %d certificate(s).", ret);
 
 	if (!tlshd_x509_client_get_certs(parms))
 		goto out_free_creds;
 	if (!tlshd_x509_client_get_privkey(parms))
-		goto out_free_creds;
+		goto out_free_certs;
 	gnutls_certificate_set_retrieve_function2(xcred,
 						  tlshd_x509_retrieve_key_cb);
 
@@ -306,7 +316,7 @@ static void tlshd_tls13_client_x509_handshake(struct tlshd_handshake_parms *parm
 	ret = gnutls_init(&session, flags);
 	if (ret != GNUTLS_E_SUCCESS) {
 		tlshd_log_gnutls_error(ret);
-		goto out_free_creds;
+		goto out_free_certs;
 	}
 	gnutls_transport_set_int(session, parms->sockfd);
 	gnutls_session_set_ptr(session, parms);
@@ -315,17 +325,17 @@ static void tlshd_tls13_client_x509_handshake(struct tlshd_handshake_parms *parm
 				     parms->peername, strlen(parms->peername));
 	if (ret != GNUTLS_E_SUCCESS) {
 		tlshd_log_gnutls_error(ret);
-		goto out_free_creds;
+		goto out_free_certs;
 	}
 	ret = gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE, xcred);
 	if (ret != GNUTLS_E_SUCCESS) {
 		tlshd_log_gnutls_error(ret);
-		goto out_free_creds;
+		goto out_free_certs;
 	}
 	ret = tlshd_gnutls_priority_set(session, parms, 0);
 	if (ret != GNUTLS_E_SUCCESS) {
 		tlshd_log_gnutls_error(ret);
-		goto out_free_creds;
+		goto out_free_certs;
 	}
 	gnutls_certificate_set_verify_function(xcred,
 					       tlshd_tls13_client_x509_verify_function);
@@ -334,6 +344,10 @@ static void tlshd_tls13_client_x509_handshake(struct tlshd_handshake_parms *parm
 
 	gnutls_deinit(session);
 
+out_free_certs:
+	tlshd_x509_client_put_privkey();
+	tlshd_x509_client_put_certs();
+
 out_free_creds:
 	gnutls_certificate_free_credentials(xcred);
 }
@@ -417,11 +431,9 @@ static void tlshd_tls13_client_psk_handshake_one(struct tlshd_handshake_parms *p
 
 	tlshd_log_debug("start ClientHello handshake");
 	tlshd_start_tls_handshake(session, parms);
-	if (!parms->session_status) {
+	if (!parms->session_status)
 		/* PSK uses the same identity for both client and server */
-		parms->num_remote_peerids = 1;
-		parms->remote_peerid[0] = peerid;
-	}
+		g_array_append_val(parms->remote_peerids, peerid);
 
 	gnutls_deinit(session);
 
@@ -432,9 +444,10 @@ out_free_creds:
 
 static void tlshd_tls13_client_psk_handshake(struct tlshd_handshake_parms *parms)
 {
-	unsigned int i;
+	key_serial_t peerid;
+	guint i;
 
-	if (!parms->peerids) {
+	if (parms->peerids->len == 0) {
 		tlshd_log_error("No key identities");
 		return;
 	}
@@ -443,8 +456,9 @@ static void tlshd_tls13_client_psk_handshake(struct tlshd_handshake_parms *parms
 	 * GnuTLS does not yet support multiple offered PskIdentities.
 	 * Retry ClientHello with each identity on the kernel's list.
 	 */
-	for (i = 0; i < parms->num_peerids; i++) {
-		tlshd_tls13_client_psk_handshake_one(parms, parms->peerids[i]);
+	for (i = 0; i < parms->peerids->len; i++) {
+		peerid = g_array_index(parms->peerids, key_serial_t, i);
+		tlshd_tls13_client_psk_handshake_one(parms, peerid);
 		if (parms->session_status != EACCES)
 			break;
 	}
@@ -489,7 +503,6 @@ static int tlshd_quic_client_set_x509_session(struct tlshd_quic_conn *conn)
 	gnutls_certificate_credentials_t cred;
 	gnutls_session_t session;
 	int ret = -EINVAL;
-	char *cafile;
 
 	if (conn->cert_req != TLSHD_QUIC_NO_CERT_AUTH) {
 		if (!tlshd_x509_client_get_certs(parms) || !tlshd_x509_client_get_privkey(parms)) {
@@ -500,14 +513,9 @@ static int tlshd_quic_client_set_x509_session(struct tlshd_quic_conn *conn)
 	ret = gnutls_certificate_allocate_credentials(&cred);
 	if (ret)
 		goto err;
-	if (tlshd_config_get_client_truststore(&cafile)) {
-		ret = gnutls_certificate_set_x509_trust_file(cred, cafile, GNUTLS_X509_FMT_PEM);
-		free(cafile);
-	} else
-		ret = gnutls_certificate_set_x509_system_trust(cred);
-	if (ret < 0)
+	ret = tlshd_client_get_truststore(cred);
+	if (ret != GNUTLS_E_SUCCESS)
 		goto err_cred;
-	tlshd_log_debug("System trust: Loaded %d certificate(s).", ret);
 
 	if (conn->cert_req == TLSHD_QUIC_NO_CERT_AUTH) {
 		gnutls_certificate_set_verify_flags(cred, GNUTLS_VERIFY_ALLOW_SIGN_RSA_MD2 |
@@ -546,6 +554,8 @@ err_session:
 err_cred:
 	gnutls_certificate_free_credentials(cred);
 err:
+	tlshd_x509_client_put_privkey();
+	tlshd_x509_client_put_certs();
 	tlshd_log_gnutls_error(ret);
 	return ret;
 }
@@ -558,7 +568,7 @@ static int tlshd_quic_client_set_anon_session(struct tlshd_quic_conn *conn)
 
 static int tlshd_quic_client_set_psk_session(struct tlshd_quic_conn *conn)
 {
-	key_serial_t peerid = conn->parms->peerids[0];
+	key_serial_t peerid = g_array_index(conn->parms->peerids, key_serial_t, 0);
 	gnutls_psk_client_credentials_t cred;
 	gnutls_session_t session;
 	char *identity = NULL;
