@@ -93,6 +93,7 @@ void tlshd_start_tls_handshake(gnutls_session_t session,
 		/* Any errors here should default to blocking access: */
 		parms->session_status = EACCES;
 		switch (ret) {
+		case GNUTLS_E_CERTIFICATE_ERROR:
 		case GNUTLS_E_CERTIFICATE_VERIFICATION_ERROR:
 			tlshd_log_cert_verification_error(session);
 			break;
@@ -181,16 +182,6 @@ void tlshd_service_socket(void)
 
 out:
 	tlshd_genl_done(&parms);
-
-	if (parms.keyring)
-		keyctl_unlink(parms.keyring, KEY_SPEC_SESSION_KEYRING);
-
-	free(parms.peerids);
-
-	if (parms.session_status) {
-		tlshd_log_failure(parms.peername, parms.peeraddr,
-				  parms.peeraddr_len);
-		return;
-	}
-	tlshd_log_success(parms.peername, parms.peeraddr, parms.peeraddr_len);
+	tlshd_log_completion(&parms);
+	tlshd_genl_put_handshake_parms(&parms);
 }
