@@ -256,24 +256,21 @@ int tlshd_keyring_link_session(const char *keyring)
 	key_serial_t serial;
 	long ret;
 
-	if (!keyring) {
+	if (!keyring || keyring[0] == '\0') {
 		tlshd_log_error("No keyring specified");
-		errno = -EINVAL;
 		return -1;
 	}
 
 	serial = find_key_by_type_and_desc("keyring", keyring, 0);
-	if (!serial) {
-		tlshd_log_debug("Failed to lookup keyring '%s'\n",
-				keyring);
-		errno = -ENOKEY;
+	if (serial == -1) {
+		tlshd_log_debug("Failed to find keyring '%s'\n", keyring);
 		return -1;
 	}
 
 	ret = keyctl_link(serial, KEY_SPEC_SESSION_KEYRING);
 	if (ret < 0) {
-		tlshd_log_debug("Failed to link keyring %s (%lx) error %d\n",
-				keyring, serial, errno);
+		tlshd_log_debug("Failed to link keyring '%s' (%lx): %s\n",
+				keyring, serial, strerror(errno));
 		return -1;
 	}
 
