@@ -104,6 +104,8 @@ tlshd_accept_nl_policy[HANDSHAKE_A_ACCEPT_MAX + 1] = {
 	[HANDSHAKE_A_ACCEPT_KEYRING]		= { .type = NLA_U32, },
 };
 
+static struct nl_sock *tlshd_notification_nls;
+
 static int tlshd_genl_event_handler(struct nl_msg *msg,
 				    __attribute__ ((unused)) void *arg)
 {
@@ -125,7 +127,13 @@ static int tlshd_genl_event_handler(struct nl_msg *msg,
 
 	if (!fork()) {
 		/* child */
+		nlmsg_free(msg);
+		tlshd_genl_sock_close(tlshd_notification_nls);
+
 		tlshd_service_socket();
+
+		tlshd_gnutls_priority_deinit();
+		tlshd_config_shutdown();
 		exit(EXIT_SUCCESS);
 	}
 
@@ -138,24 +146,23 @@ static int tlshd_genl_event_handler(struct nl_msg *msg,
  */
 void tlshd_genl_dispatch(void)
 {
-	struct nl_sock *nls;
 	int err, mcgrp;
 
-	err = tlshd_genl_sock_open(&nls);
+	err = tlshd_genl_sock_open(&tlshd_notification_nls);
 	if (err)
 		return;
 
-	nl_socket_modify_cb(nls, NL_CB_VALID, NL_CB_CUSTOM,
+	nl_socket_modify_cb(tlshd_notification_nls, NL_CB_VALID, NL_CB_CUSTOM,
 			    tlshd_genl_event_handler, NULL);
 
 	/* Let the kernel know we are running */
-	mcgrp = genl_ctrl_resolve_grp(nls, HANDSHAKE_FAMILY_NAME,
+	mcgrp = genl_ctrl_resolve_grp(tlshd_notification_nls, HANDSHAKE_FAMILY_NAME,
 				      HANDSHAKE_MCGRP_TLSHD);
 	if (mcgrp < 0) {
 		tlshd_log_error("Kernel handshake service is not available");
 		goto out_close;
 	}
-	err = nl_socket_add_membership(nls, mcgrp);
+	err = nl_socket_add_membership(tlshd_notification_nls, mcgrp);
 	if (err != NLE_SUCCESS) {
 		tlshd_log_nl_error("nl_socket_add_membership", err);
 		goto out_close;
@@ -166,9 +173,9 @@ void tlshd_genl_dispatch(void)
 		goto out_close;
 	}
 
-	nl_socket_disable_seq_check(nls);
+	nl_socket_disable_seq_check(tlshd_notification_nls);
 	while (true) {
-		err = nl_recvmsgs_default(nls);
+		err = nl_recvmsgs_default(tlshd_notification_nls);
 		if (err < 0) {
 			tlshd_log_nl_error("nl_recvmsgs", err);
 			break;
@@ -176,26 +183,21 @@ void tlshd_genl_dispatch(void)
 	};
 
 out_close:
-	tlshd_genl_sock_close(nls);
+	tlshd_genl_sock_close(tlshd_notification_nls);
 }
 
 static void tlshd_parse_peer_identity(struct tlshd_handshake_parms *parms,
 				      struct nlattr *head)
 {
+	key_serial_t peerid;
+
 	if (!head) {
 		tlshd_log_debug("No peer identities found\n");
 		return;
 	}
 
-	parms->num_peerids = 1;
-
-	parms->peerids = calloc(parms->num_peerids, sizeof(key_serial_t));
-	if (!parms->peerids) {
-		parms->num_peerids = 0;
-		return;
-	}
-
-	parms->peerids[0] = nla_get_s32(head);
+	peerid = nla_get_s32(head);
+	g_array_append_val(parms->peerids, peerid);
 }
 
 #if LIBNL_VER_NUM >= LIBNL_VER(3,5)
@@ -230,15 +232,14 @@ static void tlshd_parse_certificate(struct tlshd_handshake_parms *parms,
 		parms->x509_privkey = nla_get_s32(tb[HANDSHAKE_A_X509_PRIVKEY]);
 }
 
-static char tlshd_peername[NI_MAXHOST] = "unknown";
-static struct sockaddr_storage tlshd_peeraddr = { 0 };
-
 static int tlshd_genl_valid_handler(struct nl_msg *msg, void *arg)
 {
 	struct nlattr *tb[HANDSHAKE_A_ACCEPT_MAX + 1];
 	struct tlshd_handshake_parms *parms = arg;
+	struct sockaddr_storage addr;
+	struct sockaddr *sap = NULL;
+	socklen_t salen, optlen;
 	char *peername = NULL;
-	socklen_t optlen;
 	int err;
 
 	tlshd_log_debug("Parsing a valid netlink message\n");
@@ -251,18 +252,32 @@ static int tlshd_genl_valid_handler(struct nl_msg *msg, void *arg)
 	}
 
 	if (tb[HANDSHAKE_A_ACCEPT_SOCKFD]) {
+		char buf[NI_MAXHOST];
+		int proto;
+
 		parms->sockfd = nla_get_s32(tb[HANDSHAKE_A_ACCEPT_SOCKFD]);
-		if (getpeername(parms->sockfd, parms->peeraddr,
-				&parms->peeraddr_len) == -1) {
+
+		salen = sizeof(addr);
+		sap = (struct sockaddr *)&addr;
+		if (getpeername(parms->sockfd, sap, &salen) == -1) {
 			tlshd_log_perror("getpeername");
 			return NL_STOP;
 		}
-		optlen = sizeof(parms->ip_proto);
+		err = getnameinfo(sap, salen, buf, sizeof(buf),
+				  NULL, 0, NI_NUMERICHOST);
+		if (err) {
+			tlshd_log_gai_error(err);
+			return NL_STOP;
+		}
+		parms->peeraddr = strdup(buf);
+
+		optlen = sizeof(proto);
 		if (getsockopt(parms->sockfd, SOL_SOCKET, SO_PROTOCOL,
-			       &parms->ip_proto, &optlen) == -1) {
+			       &proto, &optlen) == -1) {
 			tlshd_log_perror("getsockopt (SO_PROTOCOL)");
 			return NL_STOP;
 		}
+		parms->ip_proto = proto;
 	}
 	if (tb[HANDSHAKE_A_ACCEPT_MESSAGE_TYPE])
 		parms->handshake_type = nla_get_u32(tb[HANDSHAKE_A_ACCEPT_MESSAGE_TYPE]);
@@ -287,24 +302,25 @@ static int tlshd_genl_valid_handler(struct nl_msg *msg, void *arg)
 	tlshd_parse_certificate(parms, tb[HANDSHAKE_A_ACCEPT_CERTIFICATE]);
 
 	if (peername)
-		strncpy(tlshd_peername, peername, sizeof(tlshd_peername) - 1);
-	else {
-		err = getnameinfo(parms->peeraddr, parms->peeraddr_len,
-				  tlshd_peername, sizeof(tlshd_peername),
+		parms->peername = strdup(peername);
+	else if (sap) {
+		char buf[NI_MAXHOST];
+
+		err = getnameinfo(sap, salen, buf, sizeof(buf),
 				  NULL, 0, NI_NAMEREQD);
 		if (err) {
 			tlshd_log_gai_error(err);
 			return NL_STOP;
 		}
+		parms->peername = strdup(buf);
 	}
 
 	return NL_SKIP;
 }
 
 static const struct tlshd_handshake_parms tlshd_default_handshake_parms = {
-	.peername		= tlshd_peername,
-	.peeraddr		= (struct sockaddr *)&tlshd_peeraddr,
-	.peeraddr_len		= sizeof(tlshd_peeraddr),
+	.peername		= NULL,
+	.peeraddr		= NULL,
 	.sockfd			= -1,
 	.ip_proto		= -1,
 	.handshake_type		= HANDSHAKE_MSG_TYPE_UNSPEC,
@@ -313,11 +329,9 @@ static const struct tlshd_handshake_parms tlshd_default_handshake_parms = {
 	.x509_cert		= TLS_NO_CERT,
 	.x509_privkey		= TLS_NO_PRIVKEY,
 	.peerids		= NULL,
-	.num_peerids		= 0,
+	.remote_peerids		= NULL,
 	.msg_status		= 0,
 	.session_status		= EIO,
-
-	.num_remote_peerids	= 0,
 };
 
 /**
@@ -340,6 +354,9 @@ int tlshd_genl_get_handshake_parms(struct tlshd_handshake_parms *parms)
 
 	*parms = tlshd_default_handshake_parms;
 
+	parms->peerids = g_array_new(FALSE, FALSE, sizeof(key_serial_t));
+	parms->remote_peerids = g_array_new(FALSE, FALSE, sizeof(key_serial_t));
+
 	ret = tlshd_genl_sock_open(&nls);
 	if (ret)
 		return ret;
@@ -398,15 +415,31 @@ out_close:
 	return ret;
 }
 
+/**
+ * tlshd_genl_put_handshake_parms - Release handshake resources
+ * @parms: handshake parameters to be released
+ *
+ */
+void tlshd_genl_put_handshake_parms(struct tlshd_handshake_parms *parms)
+{
+	if (parms->keyring)
+		keyctl_unlink(parms->keyring, KEY_SPEC_SESSION_KEYRING);
+	g_array_free(parms->peerids, TRUE);
+	g_array_free(parms->remote_peerids, TRUE);
+	free(parms->peername);
+	free(parms->peeraddr);
+}
+
 static int tlshd_genl_put_remote_peerids(struct nl_msg *msg,
 					 struct tlshd_handshake_parms *parms)
 {
-	unsigned int i;
+	key_serial_t peerid;
+	guint i;
 	int err;
 
-	for (i = 0; i < parms->num_remote_peerids; i++) {
-		err = nla_put_s32(msg, HANDSHAKE_A_DONE_REMOTE_AUTH,
-				  parms->remote_peerid[i]);
+	for (i = 0; i < parms->remote_peerids->len; i++) {
+		peerid = g_array_index(parms->remote_peerids, key_serial_t, i);
+		err = nla_put_s32(msg, HANDSHAKE_A_DONE_REMOTE_AUTH, peerid);
 		if (err < 0) {
 			tlshd_log_nl_error("nla_put peer id", err);
 			return -1;
@@ -416,7 +449,7 @@ static int tlshd_genl_put_remote_peerids(struct nl_msg *msg,
 }
 
 /**
- * tlshd_genl_done - Indicate anon handshake has completed successfully
+ * tlshd_genl_done - Indicate handshake has completed successfully
  * @parms: buffer filled in with parameters
  *
  */
