@@ -46,9 +46,6 @@ static gnutls_privkey_t tlshd_server_privkey;
 static unsigned int tlshd_server_certs_len = TLSHD_MAX_CERTS;
 static gnutls_pcert_st tlshd_server_certs[TLSHD_MAX_CERTS];
 
-/*
- * XXX: After this point, tlshd_server_certs should be deinited on error.
- */
 static bool tlshd_x509_server_get_certs(struct tlshd_handshake_parms *parms)
 {
 	if (parms->x509_cert != TLS_NO_CERT)
@@ -59,9 +56,14 @@ static bool tlshd_x509_server_get_certs(struct tlshd_handshake_parms *parms)
 					     &tlshd_server_certs_len);
 }
 
-/*
- * XXX: After this point, tlshd_server_privkey should be deinited on error.
- */
+static void tlshd_x509_server_put_certs(void)
+{
+	unsigned int i;
+
+	for (i = 0; i < tlshd_server_certs_len; i++)
+		gnutls_pcert_deinit(&tlshd_server_certs[i]);
+}
+
 static bool tlshd_x509_server_get_privkey(struct tlshd_handshake_parms *parms)
 {
 	if (parms->x509_privkey != TLS_NO_PRIVKEY)
@@ -70,6 +72,11 @@ static bool tlshd_x509_server_get_privkey(struct tlshd_handshake_parms *parms)
 	return tlshd_config_get_server_privkey(&tlshd_server_privkey);
 }
 
+static void tlshd_x509_server_put_privkey(void)
+{
+	gnutls_privkey_deinit(tlshd_server_privkey);
+}
+
 static void tlshd_x509_log_issuers(const gnutls_datum_t *req_ca_rdn, int nreqs)
 {
 	char issuer_dn[256];
@@ -126,6 +133,35 @@ tlshd_x509_retrieve_key_cb(gnutls_session_t session,
 	return 0;
 }
 
+static int tlshd_server_get_truststore(gnutls_certificate_credentials_t cred)
+{
+	char *pathname;
+	int ret;
+
+	if (tlshd_config_get_server_truststore(&pathname)) {
+		ret = gnutls_certificate_set_x509_trust_file(cred, pathname,
+							     GNUTLS_X509_FMT_PEM);
+		free(pathname);
+	} else
+		ret = gnutls_certificate_set_x509_system_trust(cred);
+	if (ret < 0)
+		return ret;
+	tlshd_log_debug("System trust: Loaded %d certificate(s).", ret);
+
+	if (tlshd_config_get_server_crl(&pathname)) {
+		ret = gnutls_certificate_set_x509_crl_file(cred, pathname,
+							   GNUTLS_X509_FMT_PEM);
+		free(pathname);
+		if (ret < 0 ) {
+			tlshd_log_gnutls_error(ret);
+			return ret;
+		}
+		tlshd_log_debug("System CRL: Loaded %d CRL(s).", ret);
+	}
+
+	return GNUTLS_E_SUCCESS;
+}
+
 /**
  * tlshd_server_x509_verify_function - Verify remote's x.509 certificate
  * @session: session in the midst of a handshake
@@ -145,9 +181,7 @@ static int tlshd_server_x509_verify_function(gnutls_session_t session,
 					     struct tlshd_handshake_parms *parms)
 {
 	const gnutls_datum_t *peercerts;
-	gnutls_certificate_type_t type;
-	unsigned int i, status;
-	gnutls_datum_t out;
+	unsigned int i, status, num_peercerts;
 	int ret;
 
 	ret = gnutls_certificate_verify_peers3(session, NULL, &status);
@@ -161,10 +195,6 @@ static int tlshd_server_x509_verify_function(gnutls_session_t session,
 		tlshd_log_gnutls_error(ret);
 		goto certificate_error;
 	}
-	type = gnutls_certificate_type_get(session);
-	gnutls_certificate_verification_status_print(status, type, &out, 0);
-	tlshd_log_debug("%s", out.data);
-	gnutls_free(out.data);
 	if (status)
 		goto certificate_error;
 
@@ -172,20 +202,18 @@ static int tlshd_server_x509_verify_function(gnutls_session_t session,
 	 * to get picky. Kernel would have to tell us what to look for
 	 * via a netlink attribute. */
 
-	peercerts = gnutls_certificate_get_peers(session,
-						 &parms->num_remote_peerids);
-	if (!peercerts || parms->num_remote_peerids == 0) {
+	peercerts = gnutls_certificate_get_peers(session, &num_peercerts);
+	if (!peercerts || num_peercerts == 0) {
 		tlshd_log_debug("The peer cert list is empty.");
 		goto certificate_error;
 	}
 
 	tlshd_log_debug("The peer offered %u certificate(s).",
-			parms->num_remote_peerids);
+			num_peercerts);
 
-	if (parms->num_remote_peerids > ARRAY_SIZE(parms->remote_peerid))
-		parms->num_remote_peerids = ARRAY_SIZE(parms->remote_peerid);
-	for (i = 0; i < parms->num_remote_peerids; i++) {
+	for (i = 0; i < num_peercerts; i++) {
 		gnutls_x509_crt_t cert;
+		key_serial_t peerid;
 
 		gnutls_x509_crt_init(&cert);
 		ret = gnutls_x509_crt_import(cert, &peercerts[i],
@@ -195,8 +223,8 @@ static int tlshd_server_x509_verify_function(gnutls_session_t session,
 			gnutls_x509_crt_deinit(cert);
 			return GNUTLS_E_CERTIFICATE_ERROR;
 		}
-		parms->remote_peerid[i] =
-			tlshd_keyring_create_cert(cert, parms->peername);
+		peerid = tlshd_keyring_create_cert(cert, parms->peername);
+		g_array_append_val(parms->remote_peerids, peerid);
 		gnutls_x509_crt_deinit(cert);
 	}
 
@@ -218,7 +246,6 @@ static void tlshd_tls13_server_x509_handshake(struct tlshd_handshake_parms *parm
 {
 	gnutls_certificate_credentials_t xcred;
 	gnutls_session_t session;
-	char *cafile;
 	int ret;
 
 	ret = gnutls_certificate_allocate_credentials(&xcred);
@@ -226,32 +253,21 @@ static void tlshd_tls13_server_x509_handshake(struct tlshd_handshake_parms *parm
 		tlshd_log_gnutls_error(ret);
 		return;
 	}
+	ret = tlshd_server_get_truststore(xcred);
+	if (ret != GNUTLS_E_SUCCESS)
+		goto out_free_creds;
 
-	if (tlshd_config_get_server_truststore(&cafile)) {
-		ret = gnutls_certificate_set_x509_trust_file(xcred, cafile,
-							     GNUTLS_X509_FMT_PEM);
-		free(cafile);
-	} else
-		ret = gnutls_certificate_set_x509_system_trust(xcred);
-	if (ret < 0) {
-		tlshd_log_gnutls_error(ret);
+	if (!tlshd_x509_server_get_certs(parms))
 		goto out_free_creds;
-	}
-	tlshd_log_debug("System trust: Loaded %d certificate(s).", ret);
-
-	if (!tlshd_x509_server_get_certs(parms)) {
-		goto out_free_creds;
-	}
-	if (!tlshd_x509_server_get_privkey(parms)) {
-		goto out_free_creds;
-	}
+	if (!tlshd_x509_server_get_privkey(parms))
+		goto out_free_certs;
 	gnutls_certificate_set_retrieve_function2(xcred,
 						  tlshd_x509_retrieve_key_cb);
 
 	ret = gnutls_init(&session, GNUTLS_SERVER);
 	if (ret != GNUTLS_E_SUCCESS) {
 		tlshd_log_gnutls_error(ret);
-		goto out_free_creds;
+		goto out_free_certs;
 	}
 	gnutls_transport_set_int(session, parms->sockfd);
 	gnutls_session_set_ptr(session, parms);
@@ -259,7 +275,7 @@ static void tlshd_tls13_server_x509_handshake(struct tlshd_handshake_parms *parm
 	ret = gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE, xcred);
 	if (ret != GNUTLS_E_SUCCESS) {
 		tlshd_log_gnutls_error(ret);
-		goto out_free_creds;
+		goto out_free_certs;
 	}
 	gnutls_certificate_set_verify_function(xcred,
 					       tlshd_tls13_server_x509_verify_function);
@@ -268,13 +284,39 @@ static void tlshd_tls13_server_x509_handshake(struct tlshd_handshake_parms *parm
 	ret = tlshd_gnutls_priority_set(session, parms, 0);
 	if (ret) {
 		tlshd_log_gnutls_error(ret);
-		goto out_free_creds;
+		goto out_free_certs;
 	}
 
 	tlshd_start_tls_handshake(session, parms);
 
+	if (tlshd_debug &&
+	    gnutls_certificate_type_get(session) == GNUTLS_CRT_X509) {
+		const gnutls_datum_t *peercerts;
+		unsigned int i, num_certs = 0;
+
+		peercerts = gnutls_certificate_get_peers(session, &num_certs);
+		for (i = 0; i < num_certs; i++) {
+			gnutls_x509_crt_t cert;
+			gnutls_datum_t cinfo;
+
+			gnutls_x509_crt_init(&cert);
+			gnutls_x509_crt_import(cert, &peercerts[i],
+					       GNUTLS_X509_FMT_DER);
+			if (gnutls_x509_crt_print(cert, GNUTLS_CRT_PRINT_ONELINE,
+						  &cinfo) == 0) {
+				tlshd_log_debug("Peer certificate: %s", cinfo.data);
+				gnutls_free(cinfo.data);
+			}
+			gnutls_x509_crt_deinit(cert);
+		}
+	}
+
 	gnutls_deinit(session);
 
+out_free_certs:
+	tlshd_x509_server_put_privkey();
+	tlshd_x509_server_put_certs();
+
 out_free_creds:
 	gnutls_certificate_free_credentials(xcred);
 }
@@ -316,8 +358,7 @@ static int tlshd_server_psk_cb(gnutls_session_t session,
 	}
 
 	/* PSK uses the same identity for both client and server */
-	parms->remote_peerid[0] = psk;
-	parms->num_remote_peerids = 1;
+	g_array_append_val(parms->remote_peerids, psk);
 	return 0;
 }
 
@@ -453,8 +494,7 @@ found:
 	}
 
 	/* PSK uses the same identity for both client and server */
-	parms->remote_peerid[0] = psk;
-	parms->num_remote_peerids = 1;
+	g_array_append_val(parms->remote_peerids, psk);
 	return 0;
 }
 
@@ -465,7 +505,6 @@ static int tlshd_quic_server_set_x509_session(struct tlshd_quic_conn *conn)
 	gnutls_datum_t ticket_key;
 	gnutls_session_t session;
 	int ret = -EINVAL;
-	char *cafile;
 
 	if (!tlshd_x509_server_get_certs(parms) || !tlshd_x509_server_get_privkey(parms)) {
 		tlshd_log_error("cert/privkey get error %d", -ret);
@@ -475,15 +514,9 @@ static int tlshd_quic_server_set_x509_session(struct tlshd_quic_conn *conn)
 	ret = gnutls_certificate_allocate_credentials(&cred);
 	if (ret)
 		goto err;
-	if (tlshd_config_get_server_truststore(&cafile)) {
-		ret = gnutls_certificate_set_x509_trust_file(cred, cafile,
-							     GNUTLS_X509_FMT_PEM);
-		free(cafile);
-	} else
-		ret = gnutls_certificate_set_x509_system_trust(cred);
-	if (ret < 0)
-		goto err_cred;
-	tlshd_log_debug("System trust: Loaded %d certificate(s).", ret);
+	ret = tlshd_server_get_truststore(cred);
+	if (ret)
+		goto err;
 
 	gnutls_certificate_set_retrieve_function2(cred, tlshd_x509_retrieve_key_cb);
 
@@ -528,6 +561,8 @@ err_session:
 err_cred:
 	gnutls_certificate_free_credentials(cred);
 err:
+	tlshd_x509_server_put_privkey();
+	tlshd_x509_server_put_certs();
 	tlshd_log_gnutls_error(ret);
 	return ret;
 }
